
/* ****************************************************
 * project	: bootloader slave
 * 		:
 * platform	: avr-gcc-4.1.2
 * 		: avr-binutils-2.17
 * 		: avr-libc-1.4.8
 * 		:
 * pc os	: Fedora8
 * editor	: VIM(utf-8)
 *		:
 * author	: mowin
 * date		: 2008-06-15
 * ***************************************************/

/* ***************************************************
 * 08-07-04	: ver0.03
 * 		: 优化代码大小
 * 		: 加入支持多种页大小不同的MCU
 * 		: 已经测试通过: M64,M16,M8
 * 		:
 * 08-07-03	: ver0.02
 * 		: 清除使用过的定时器寄存器，以使用户程序
 * 		: 不因此而出现运行异常
 * 		: 
 * 2008-06-15	: ver0.01
 * 		: 原始版本
 * 		: 只测试了M16
 * **************************************************/

/* 只包含需要的头文件 */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "config.h"
#include "app.h"

#include "lcd/lcd.h"

/* XMODEM 数据帧 */
#define PACK_NO1   0
#define PACK_NO2   1
#define DATA_START 2
#define DATA_LEN   128
#define CRCH	   130
#define CRCL       131
#define XMODEM_FRAME_LEN 132

#define CRC_EN 1 /* 1 启用CRC 0 禁用CRC */

#if CRC_EN
/* 计算CRC校验(1021) */
uint8_t ch, cl;
void crc16(uint8_t *buf, uint8_t n)
{
    uint8_t j;
    uint8_t i;
    uint16_t crc, t;

    crc = 0;
    for(j = n; j > 0; j--){
	crc = (crc ^ (((uint16_t) *buf) << 8));
	for(i = 8; i > 0; i--){
	    t = crc << 1;
	    if(crc & 0x8000)
		t = t ^ 0x1021;
	    crc = t;
	}
	buf++;
    }
    ch = crc / 256;
    cl = crc % 256;
}
#endif

void turn_off_timer1(void)
{
    TCCR1A = 0;
    TCCR1B = 0;
    TCNT1 = 0;
    OCR1A = 0;
    OCR1B = 0;
}

/* *********************************
 * function 	: main
 * argument(s) 	: void
 * return	: 0=success 1=fail
 * ********************************/
int main(void)
{
#define TRUE 0
#define FALSE -1

#if FLASHEND > 0xffff /* flash > 64K */
    uint32_t flash_addr = 0;
#else
    uint16_t flash_addr = 0;
#endif

    /* receive buffer */
    uint8_t buf[XMODEM_FRAME_LEN];

    /* pack number start from 1 */
    uint8_t pack_no = 1;

    /* counter */
    uint8_t cnt = 0;

    /* loop counter */
    uint8_t i;

#if SPM_PAGESIZE < DATA_LEN /* 一个数据包写入多页(M48,M8) */
    uint8_t page_write_ptr = 0;
#elif SPM_PAGESIZE > DATA_LEN /* 多个数据包写入一页(M64) */
    uint8_t page_buf[SPM_PAGESIZE];
    uint16_t copy_cnt = 0;
#else
    /* 一个包写入一页(M16,M32) */
#endif


    cli();
    device_init();

    lcd_init();
    disp();

#if TRIGGER_MODE == 1
    /* 串口触发 */
    /* waiting for connected code 'd' */
    cnt = WAIT_C_CNT;
    while(1){

	if(CHKBIT(TIFR, OCF1A)){
	    TIFR = _BV(OCF1A);/* write 1 clear flag */

#if LED_EN
	    LED_ACT;
#endif
	    if(--cnt == 0){
		turn_off_timer1();
		quit();
	    }
	}

	if(CHECK_RXC)
	    if(UDR_REG == 'd')
		break; /* receive connected code */
    }
#elif TRIGGER_MODE == 0
    /* 
     * 端口触发
     * TRI_PIN = 0 --> bootloader
     * TRI_PIN = 1 --> normal boot 
     */
    cnt = WAIT_C_CNT;
    while(CHECK_TRI_PIN){

	if(CHKBIT(TIFR, OCF1A)){
	    TIFR = _BV(OCF1A);/* write 1 clear flag */

#if LED_EN
	    LED_ACT;
#endif
	    if(--cnt == 0){
		turn_off_timer1();
		quit();
	    }
	}
    }
#else
#endif

    /* transmit XMODEM_RWC and wait for XMODEM_SOH */
    cnt = WAIT_D_CNT;
    i = 0;
    while(1){

	/* 收到XMODEM_SOH后开始接收数据 */
	if(CHECK_RXC){
	    if(UDR_REG == XMODEM_SOH){
		break; /* start */
	    }
	}

	if(CHKBIT(TIFR, OCF1A)){
	    /* 100MS */
	    TIFR = _BV(OCF1A);/* write 1 clear flag */

	    /* 发送XMODEM＿RWC */
	    write_udr(XMODEM_RWC);
#if LED_EN
	    LED_ACT;
#endif

	    /* 10秒未收到XMODEM_SOH退出至用户程序 */
	    if(--cnt == 0){
		turn_off_timer1();
		quit();
	    }
	}
    }

    /* turn off timer */
    turn_off_timer1();

    cnt = 0;
    /* start receive data frame */
    do{
	/* 读入一帧数据 */
	for(i = 0; i < XMODEM_FRAME_LEN; i++)
	    buf[i] = read_udr();

	buf[PACK_NO2] = ~buf[PACK_NO2];
	/* 检验包序列号 */
	if(buf[PACK_NO1] == pack_no && buf[PACK_NO2] == pack_no){

#if CRC_EN
	    /* 校验 CRC */
	    crc16(&buf[DATA_START], DATA_LEN);
	    if(buf[CRCH] == ch && buf[CRCL] == cl){
#endif
		/* 数据正确 */

#if SPM_PAGESIZE == DATA_LEN
		boot_program_page(flash_addr, &buf[DATA_START]);
		flash_addr += SPM_PAGESIZE;/* 下一页FLASH */

#elif SPM_PAGESIZE < DATA_LEN
		page_write_ptr += DATA_START;/* 数据起始地址 */
		do{
		    boot_program_page(flash_addr, &buf[page_write_ptr]);
		    flash_addr += SPM_PAGESIZE; /* 下一页 */
		    page_write_ptr += SPM_PAGESIZE; /* 数据指针指向下一页 */
		} while(page_write_ptr < DATA_LEN + DATA_START);
		page_write_ptr = 0;

#elif SPM_PAGESIZE > DATA_LEN
		for(i = 0; i < DATA_LEN; i++)
		    page_buf[copy_cnt++] = buf[i + DATA_START];

		/* 等待数据填满一页缓冲之后写入 */
		if(copy_cnt >= SPM_PAGESIZE){
		    boot_program_page(flash_addr, page_buf);
		    flash_addr += SPM_PAGESIZE;
		    copy_cnt = 0;
		}
#endif

		/* 下一包 */
		pack_no++;
		write_udr(XMODEM_ACK);
		cnt = 0;
#if LED_EN
		LED_ACT;/* green LED */
#endif

#if CRC_EN
	    } else {
		/* CRC校验错重发 */
		write_udr(XMODEM_NAK);
		cnt++;
		//LED2_ACT;/* red LED */
	    }
#endif

	} else {
	    /* 包序列号错重发*/
	    write_udr(XMODEM_NAK);
	    cnt++;
	    //LED2_ACT;
	}

	/* 误码太多放弃升级 */
	//if(cnt > 10)
	//   break;


    } while(read_udr() != XMODEM_EOT);
    write_udr(XMODEM_ACK);

    quit();
    return TRUE;
}
/* end of file */
